#!/usr/bin/env python3
# coding: utf-8

## ===============================
## NOT CURRENTLY BEING USED
## ===============================

import argparse
import importlib
import logging
import numpy as np
import os
import pandas as pd
import re

from dart_id.helper import *

logger = logging.getLogger('root')

def write_outputs(df, out_folder):

  #with open(out_file, 'w') as f:

  graph_text = ''
  targets = []
  decoys = []

  prot_list = []
  prot_cols = ['Leading proteins', 'Proteins']

  for i in range(0, df.shape[0]):

    prot_list = []
    for col in prot_cols:
      prots = df[col].loc[i]
      if pd.isnull(prots): continue
      #prot_list = prot_list + [str.join('_', prot.split('|')[0:2]) for prot in prots.split(';')]
      prot_list = prot_list + [prot for prot in prots.split(';')]

    prot_list = np.unique(prot_list)
    if len(prot_list) < 1: continue

    graph_text += ('e ' + df['Sequence'].loc[i] + '\n')

    for prot in prot_list:
      graph_text += ('r ' + prot + '\n')

      if 'REV__' in prot:
        # add to decoy list
        decoys.append(prot)
      else:
        # add to targets list
        targets.append(prot)
        

    graph_text += ('p ' + str(1-df['PEP'].loc[i]) + '\n')
    
  targets = np.unique(targets)
  decoys = np.unique(decoys)

  # write to strings
  target_decoy_text = '{ '
  for i, t in enumerate(targets):  
    if i == len(targets)-1:
      target_decoy_text += (t + ' }')
    else: 
      target_decoy_text += (t + ' , ')

  target_decoy_text += ('\n{ ')
  for i, d in enumerate(decoys):
    if i == len(targets)-1:
      target_decoy_text += (d + ' }')
    else: 
      target_decoy_text += (d + ' , ')


  # write psm graph
  with open(os.path.join(out_folder, 'psm_graph.txt'), 'w') as f:
    f.write(graph_text)
    f.close()

  # write target decoy file
  with open(os.path.join(out_folder, 'target_decoy.txt'), 'w') as f:
    f.write(target_decoy_text)
    f.close()



def main():
  # load command-line args
  parser = argparse.ArgumentParser()  
  add_global_args(parser, add_config_file=False)
  #parser.add_argument()
  args = parser.parse_args()

  # initialize logger
  init_logger(args.verbose, '', log_to_file=False)
  logger = logging.getLogger('root')

  df = pd.DataFrame()

  # iterate through each input file provided.
  for i, f in enumerate(args.input):
    logger.info('Reading in input file #{} | {} ...'.format(i+1, f.name))

    # load the input file
    dfa = pd.read_csv(f.name, sep='\t', low_memory=False)
    # append a copy of dfa into df, because the conversion process will heavily
    # modify dfa. we need to keep a copy of the original dataframe in order to append
    # the new columns back onto it later.
    # re-index columns with '[dfa.columns.tolist()]' to preserve the general column order
    df = df.append(dfa)[dfa.columns.tolist()]

  df = df.sort_values(by=['Sequence']).reset_index(drop=True)

  # write psm graph file
  write_outputs(df, args.output)
  

if __name__ == '__main__':
  main()


#!/usr/bin/env python3
# coding: utf-8

import logging
import matplotlib
matplotlib.use('PS')
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from dart_id.helper import *
from scipy.stats import norm, lognorm, laplace

logger = logging.getLogger('root')

def gen(df, config, params, output_path):
  figures_path = create_fig_folder(output_path, 'figures')
  fig_names = []

  num_experiments = len(df['exp_id'].unique())
  df['residual'] = df[config['col_names']['retention_time']] - df['muij']

  plots_per_row = 30
  if num_experiments < plots_per_row:
    plots_per_row = num_experiments

  num_rows = int(np.ceil(num_experiments / plots_per_row))

  resi = []

  for i in range(0, num_rows):
      ax = plt.subplot2grid((num_rows, 1), (i, 0))
      
      if (i + 1) * plots_per_row > num_experiments:
          resi = [df['residual'][(df['exp_id'] == i) & (~pd.isnull(df['residual']))] for i in range((i * plots_per_row), num_experiments)]
          ax.boxplot(resi, showfliers=False)
          ax.set_xticklabels(np.arange((i * plots_per_row), num_experiments, 1))
      else:
          resi = [df['residual'][(df['exp_id'] == i) & (~pd.isnull(df['residual']))] for i in range((i * plots_per_row), ((i + 1) * plots_per_row))]
          ax.boxplot(resi, showfliers=False)
          ax.set_xticklabels(np.arange((i * plots_per_row), ((i + 1) * plots_per_row), 1))
          
      #ax.violinplot(resi, showmedians=True, showextrema=True)
      #ax.boxplot(resi, showfliers=False)
      ax.set_xticks(np.arange(1, plots_per_row + 1, 1))

      ax.set_xlabel('Experiment Number')
      ax.set_ylabel('Residual RT (min)')
      
  #plt.subplots_adjust(hspace=0.6, wspace=0.3)
  #plt.tight_layout()

  # finalize and save figure
  f = plt.gcf()
  #f.text(0.5, 0, 'Experiment Number', fontsize=16, ha='center', va='center')
  #f.text(0.06, 0.5, 'Residual RT (min)', fontsize=16, ha='center', va='center', rotation='vertical')
  f.set_size_inches(12, num_rows * 2)

  fname = os.path.join(figures_path, 'residuals_violin.png')
  logger.info('Saving figure to {} ...'.format(fname))
  f.savefig(fname, dpi=160)
  fig_names.append(fname)

  plt.close()
  f.clf()

  return fig_names



#!/usr/bin/env python3
# coding: utf-8

import argparse
import json
import logging
import numpy as np
import os
import pandas as pd
import pkg_resources
import platform
import re
try:
  import subprocess32 as subprocess
except ImportError:
  import subprocess
import sys
import time

from dart_id.converter import process_files
from dart_id.exceptions import STANError
from dart_id.models import models, get_model_from_config
from dart_id.helper import init_logger, add_global_args, read_config_file, convert_numpy_scalar

pd.options.mode.chained_assignment = None
logger = logging.getLogger('root')

def get_os():
  x = platform.system()
  if x == 'Linux': 
    # get the linux distro
    p = platform.platform()
    if re.search(r'centos-7', p) is not None or re.search(r'rhel', p) is not None:
      return 'rhel'
    elif re.search(r'debian', p) is not None:
      return 'debian'
    else:
      logger.warning('Linux distribution {} not recognized. Defaulting to generic linux tag...'.format(p))
      return 'linux_generic'

  elif x == 'Darwin': return 'mac'
  elif x == 'Windows': return 'windows'
  else: raise Exception('Operating system {} not identified'.format(x))

def get_exec_name(model):
  x = platform.system()
  if x == 'Windows':
    return '{}.exe'.format(model)
  else: return model

def align(dfa, config):

  # take subset of confident observations to use for alignment
  dff = dfa[-(dfa['exclude'])].reset_index(drop=True)

  #logger.info('{} / {} ({:.2%}) confident, alignable observations (PSMs) after filtering.'.format(dff.shape[0], dfa.shape[0], dff.shape[0] / dfa.shape[0]))

  # refactorize peptide id into stan_peptide_id, 
  # to preserve continuity when feeding data into STAN
  dff['stan_peptide_id'] = dff['sequence'].map({
    ind: val for val, ind in enumerate(dff['sequence'].unique())})

  exp_names = np.sort(dff['raw_file'].unique())
  num_experiments = len(exp_names)
  num_observations = dff.shape[0]
  num_peptides = dff['stan_peptide_id'].max() + 1
  
  logger.info('Building peptide-experiment pairs...')

  # build a unique peptide-experiment ID
  pep_exp_all = dff['stan_peptide_id'].map(str) + ' - ' + dff['exp_id'].map(str)
  pep_exp_pairs = pep_exp_all.unique()
  num_pep_exp_pairs = len(pep_exp_pairs)

  # muij_map - maps pair ID back to dataframe index
  muij_map = pep_exp_all.map({ind: val for val, ind in enumerate(pep_exp_pairs)})
  # maps to experiment and peptide ID
  splt = pd.Series(pep_exp_pairs).str.split(' - ')
  muij_to_pep = splt.str.get(0).map(int)
  muij_to_exp = splt.str.get(1).map(int)

  logger.info('{} peptide-experiment pairs.'.format(num_pep_exp_pairs))

  # build dictionary of data to feed to STAN
  # revert all data to primitive types to avoid problems later
  # STAN code is all 1-indexed, so add 1 to all indexed forms of data
  #
  # convert any numpy objects to base python types/lists, so they can
  # be printed in non-binary form when this dict is converted to JSON
  stan_data = {
    'num_experiments': convert_numpy_scalar(num_experiments),
    'num_peptides': convert_numpy_scalar(num_peptides),
    'num_pep_exp_pairs': convert_numpy_scalar(num_pep_exp_pairs),
    'num_total_observations': convert_numpy_scalar(num_observations),
    'muij_map': (muij_map+1).tolist(),
    'muij_to_pep': (muij_to_pep+1).tolist(),
    'muij_to_exp': (muij_to_exp+1).tolist(),
    'experiment_id': (dff['exp_id']+1).tolist(),
    'peptide_id': (dff['stan_peptide_id']+1).tolist(),
    'retention_times': dff['retention_time'].tolist(),
    'mean_log_rt': convert_numpy_scalar(np.mean(np.log(dff['retention_time']))),
    'sd_log_rt': convert_numpy_scalar(np.std(np.log(dff['retention_time']))),
    'rt_mean': convert_numpy_scalar(np.mean(dff['retention_time'])),
    'rt_std': convert_numpy_scalar(np.std(dff['retention_time'])),
    'pep': dff['pep'].tolist(),
    'max_retention_time': convert_numpy_scalar(dff['retention_time'].max())
  }

  logger.info('Initializing fit priors for {} peptides...'.format(num_peptides))

  model = get_model_from_config(config)

  logger.info('Aligning RTs using the \"{}\" model'.format(model['name']))

  # generate initial values with the given function in models.py
  init_list = model['init_func'](dff, config)

  # init sigmas (spread) for each peptide
  #sigma_init = np.zeros(num_peptides)
  #muijs = init_list['beta_0'][dff['exp_id']] + \
  #  (init_list['beta_1'][dff['exp_id']] * init_list['mu'][dff['stan_peptide_id']])
  #for i in range(0, num_peptides):
  #  sigma_init[i] = np.std(muijs[dff['stan_peptide_id']==i])

  # add sigma statistics to data list, to build sampling dist off of
  #stan_data['sigma_mean'] = np.mean(sigma_init)
  #stan_data['sigma_std'] = np.std(sigma_init)
  
  
  logger.info('Preparing for STAN Alignment')

  logger.info('Saving data as JSON to pass to STAN executable...')

  stan_input_path = os.path.join(config['output'], 'stan_input.json')
  with open(stan_input_path, 'w') as f:
    logger.info('Saving input data to stan_input.json')
    f.write(json.dumps(stan_data))
    f.close()
  
  stan_init_list_path = os.path.join(config['output'], 'stan_init_list.json')
  with open(stan_init_list_path, 'w') as f:
    logger.info('Saving initial values to stan_init_list.json')
    f.write(json.dumps(init_list))
    f.close()

  exec_folder = get_os()
  logger.info('Detected operating system: {} ({} | {})'.format(exec_folder, platform.system(), platform.platform()))
  exec_name = get_exec_name(model['model_binary'])
  
  # find the model folder in models/
  model_exists = pkg_resources.resource_exists('dart_id', '/'.join(['models', model['model_binary'], exec_folder, exec_name]))
  if not model_exists:
    raise Exception('Compiled model \"{}\" does not exist for operating system \"{}\". Please contact the authors for the executable or build it locally on your machine with cmdstan.'.format(model['model_binary'], exec_folder))

  model_path = pkg_resources.resource_filename('dart_id', '/'.join(('models', model['model_binary'], exec_folder, exec_name)))
  logger.info('Found executable for model \"{}\" in: {}'.format(model['model_name'], model_path))


  stan_output_path = os.path.join(config['output'], 'stan_output.csv')

  cmd = [
    model_path,
    'optimize', 'iter={}'.format(config['stan_iters']),
    'algorithm=lbfgs',
      'init_alpha={}'.format(config['init_alpha']),
      'tol_obj={}'.format(config['tol_obj']),
      'tol_rel_obj={}'.format(config['tol_rel_obj']),
      'tol_grad={}'.format(config['tol_grad']),
      'tol_rel_grad={}'.format(config['tol_rel_grad']),
      'tol_param={}'.format(config['tol_param']),
      'history_size={}'.format(config['history_size']),
    'init={}'.format(stan_init_list_path),
    'data', 'file={}'.format(stan_input_path),
    'output', 'file={}'.format(stan_output_path)
  ]

  # run STAN and time it
  logger.info('Running STAN with command: {}'.format(' '.join(cmd)))
  start = time.time()

  #stan_log_path = os.path.join(config['output'], 'stan_alignment.log')
  #output_fd = open(stan_log_path, 'w')
  proc = subprocess.Popen(cmd)

  while True:
    try:
      proc.wait(1)
      break
    except subprocess.TimeoutExpired:
      continue
    except KeyboardInterrupt:
      break
  
  logger.info('STAN Model Finished. Run time: {:.3f} seconds.'.format(time.time() - start))

  #output_fd.close()

  #stdout = ''
  #with open(stan_log_path, 'r') as fd:
  #  stdout = fd.read()

  if proc.returncode != 0:
    logger.warning('Stan model exited with error {}'.format(proc.returncode))
  
  # parse the output file
  with open(stan_output_path, 'r') as f:
    stan_out = f.readlines()
    # we only need the parameters from the last iteration,
    # so just read the headers and the last line
    headers = stan_out[-2]
    values = stan_out[-1]
    f.close()

  # transform into a dict
  stan_results = dict()
  headers = headers.split(',')
  values = values.split(',')
  
  start = 0
  cur_header = headers[0].split('.')[0]
  for i, header in enumerate(headers):
    header = header.split('.')[0]

    if header != cur_header or i == (len(headers) - 1):
      # where to end the slice
      end = i
      if i == (len(headers) - 1): end = i + 1
      # store value slice for the previous header
      stan_results[cur_header] = np.array(values[start:end])
      # reset counter
      start = i
      # reset previous header
      cur_header = header
    else: continue

  if len(stan_results) == 0:
    raise STANError('STAN did not return any parameters, or the returned parameters could not be read. Please check your output folder for the file \"stan_output.csv\" and verify that it exists and holds the parameter data.')

  # convert strings to ints or floats
  for param in stan_results:
    is_float = '.' in stan_results[param][0]
    if is_float:
      stan_results[param] = stan_results[param].astype(float)
    else:
      stan_results[param] = stan_results[param].astype(int)

  # exp_params - contains regression parameters for each experiment
  # peptide_params - contains canonical RT (mu) for each peptide sequence
  # pair_params - contains muij and sigmaij for each peptide-experiment pair
  #               not exactly necessary as these can be extrapolated from 
  #               exp_params and peptide_params
  
  exp_params = pd.DataFrame({ 
    key: stan_results[key] for key in model['exp_keys']})
  peptide_params = pd.DataFrame({ key: stan_results[key] for key in model['peptide_keys']})
  pair_params = pd.DataFrame({ key: stan_results[key] for key in model['pair_keys']})

  # add exp_id to exp_params
  exp_params['exp_id'] = np.sort(dff['exp_id'].unique())

  # put the maps in the pair parameters file as well
  pair_params = pair_params.assign(muij_to_pep=muij_to_pep)
  pair_params = pair_params.assign(muij_to_exp=muij_to_exp)

  # add initial values to parameters files, and prepend those column names with 'init_'
  exp_params = pd.concat([exp_params,
    pd.DataFrame({ 'init_'+key: init_list[key] for key in model['exp_keys']})],
    axis=1)
  peptide_params = pd.concat([peptide_params,
    pd.DataFrame({ 'init_'+key: init_list[key] for key in model['peptide_keys']})],
    axis=1)
  #pair_params = pd.concat([pair_params,
  #  pd.DataFrame({ 'init_'+key: init_list[key] for key in model['pair_keys']})],
  #  axis=1, ignore_index=True)

  if config['save_params']:
    # write parameters to file, so operations can be done on alignment data without
    # the entire alignment to run again
    logger.info('Writing STAN parameters to file...')
    exp_params.to_csv(os.path.join(config['output'], 'exp_params.txt'), 
      sep='\t', index=False)
    pair_params.to_csv(os.path.join(config['output'], 'pair_params.txt'), 
      sep='\t', index=True, index_label='pair_id')
    peptide_params.to_csv(os.path.join(config['output'], 'peptide_params.txt'), 
      sep='\t', index=True, index_label='peptide_id')

  # write parameters to dict for further operations
  params = {}
  params['exp'] = exp_params
  params['pair'] = pair_params
  params['peptide'] = peptide_params

  return params

def main():
  # load command-line args
  parser = argparse.ArgumentParser()  
  add_global_args(parser)
  args = parser.parse_args()

  # load config file
  # this function also creates the output folder
  config = read_config_file(args)

  # initialize logger
  init_logger(config['verbose'], os.path.join(config['output'], 'align.log'), config['log_file'])

  logger.info('Converting files and filtering PSMs')
  df, df_original = process_files(config)
  logger.info('Finished converting files and filtering PSMs.')

  logger.info('Beginning alignment procedure')
  align(df, config)
  logger.info('Alignment procedure finished')

if __name__ == '__main__':
  main()





#!/usr/bin/env python3
# coding: utf-8

import io
import os

from dart_id.version import __version__
from os import path
from setuptools import setup, find_packages, Command

name = 'dart_id'
test_name = 'dart_id-fresca'

version=__version__
test_version='2.0.4'

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name=name,
  version=version,
  description='RT Alignment and Peptide ID Confidence Updating for LC-MS/MS Data',
  long_description=long_description,
  long_description_content_type='text/markdown',
  url='https://github.com/SlavovLab/DART-ID',
  author='Albert Chen',
  author_email='chen.alb@husky.neu.edu',
  license='MIT',
  classifiers=[
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    
    'License :: OSI Approved :: MIT License',

    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',

    'Topic :: Scientific/Engineering :: Bio-Informatics'
  ],
  keywords='bayesian retention time psm peptide spectra update',
  project_urls={
    'Documentation': 'https://github.com/SlavovLab/DART-ID',
    'Source': 'https://github.com/SlavovLab/DART-ID',
    'Tracker': 'https://github.com/SlavovLab/DART-ID/issues',
    'Lab Page': 'https://web.northeastern.edu/slavovlab/'
  },
  packages=['dart_id'],
  #libraries=[],
  #setup_requires=[],
  install_requires=[
    'pyyaml>=3.12',
    'numpy>=1.14.3',
    'scipy>=1.0.0',
    'pandas>=0.22.0',
    'matplotlib>=2.1.2',
    'networkx>=2.1'
  ],
  #test_requires=[
  #  'pytest'
  #],
  extras_require={

  },
  include_package_data=True,
  # specified in MANIFEST.in instead
  #package_data={
  #  'dart_id': [
  #    'models/*',            # STAN models
  #    'figure_gen/*',        # figure generation scripts
  #    'figure_resources/*',  # figure generation resources
  #    'fido/*'               # fido scripts
  #  ]
  #},
  # data outside the package
  #data_files=data_files,
  
  entry_points={
    'console_scripts': [
      ('dart_id=dart_id.update:main'),
      ('dart_id_convert=dart_id.converter:main'),
      ('dart_id_align=dart_id.align:main'),
      ('dart_id_update=dart_id.update:main'),
      ('dart_id_figures=dart_id.figures:main')#,
#      ('dart_id_collate=dart_id.collate:main')
    ]
  }
)



# coding: utf-8

import itertools
import networkx as nx
import dart_id.fido.BinomialHeap

from dart_id.fido.Utilities import *

def elimination_edges(g, n):
  return list(itertools.combinations(g.neighbors(n), 2))
def new_elimination_edges(g, n):
  elim_edges = elimination_edges(g,n)
  added_edges = []
  for a,b in elim_edges:
    if b not in g[a]:
      added_edges.append((a,b))
  return added_edges
def number_of_added_elimination_edges(g,n):
  ### note that it is important to make sure to not count edge (a,b)
  ### different from (b,a)
  edge_set = elimination_edges(g, n)
  total_edge_count = len(edge_set)
  existing_edge_count = 0
  for a,b in edge_set:
    if b in g[a]:
      existing_edge_count += 1
  return total_edge_count - existing_edge_count

def to_moralized(g):
  moral_graph = g.to_undirected()
  for n in g.node:
    ### this is not the most efficient way to perform this or
    ### to visualize. all predecessors of n will form a
    ### complete graph. this graph can be represented as a
    ### group containing all of these nodes.
    p_list = list(g.predecessors(n))
    for pair in itertools.combinations( p_list,  2):
      moral_graph.add_edge(*pair)
  return moral_graph

def to_triangulated(g):
  ### this method is O(n^2) and can be made more efficient with a
  ### heap. triangulating a graph consisting of two disjoint graphs is
  ### the same as triangulating each individually, so do that for now.
  return fast_to_triangulated(g)

  triangulated = nx.Graph()
  subgraphs = nx.connected_component_subgraphs(g)
  print('\t\t\ttriangulating subgraphs...')
  triangulated_subgraphs = [connected_to_triangulated(sg) for sg in subgraphs]
  print('\t\t\tassembling subgraphs into graph union...')
  triangulated = all_graph_unions(triangulated_subgraphs)
  return triangulated

def fast_to_triangulated(g):
  ### this method is O(n^2) and can be made more efficient with a
  ### heap. triangulating a graph consisting of two disjoint graphs is
  ### the same as triangulating each individually, so do that for now.
  triangulated = nx.Graph()
  subgraphs = nx.connected_component_subgraphs(g)
  #print('\tlooking into subgraph triangulation:')
  #for sg in subgraphs:
  #  print('\t\tsubgraph size:',len(sg))
  #print('\t\t\ttriangulating subgraphs...')
  triangulated_subgraphs = [fast_connected_to_triangulated(sg) for sg in subgraphs]
  #print('\t\t\tassembling subgraphs into graph union...')
  triangulated = all_graph_unions(triangulated_subgraphs)
  return triangulated

### does minimum neighbors
def fast_connected_to_triangulated(g):
  added_edges = []
  local = g.copy()
  
  nodes_and_number_of_neighbors = [ (len(local[n]),n) for n in local ]
  #print(str(nodes_and_number_of_neighbors))
  nodes_to_heap_references = {}
  bh = BinomialHeap.heap()
  for number_neighbors,n in nodes_and_number_of_neighbors:
    #print 'heap inserting',number_neighbors,n
    nodes_to_heap_references[n] = bh.insert(number_neighbors,n)
  for removal_index, n in enumerate(bh):
    #print('\t\tremoving',n,'with',nodes_to_heap_references[n].ref.key,'neighbors')
     
    # if removal_index == 1:
    #     for n in bh:
    #         print '\t\t\tresult:', n, 'with',nodes_to_heap_references[n].ref.key,'neighbors'
    #     nx.draw_graphviz(local)
    #     P.show()

    edges = new_elimination_edges(local, n)

    ### count the edges actually added incident to each node
    nodes_to_added_edges = counting_dict( [ (nei,0) for nei in local.neighbors(n) ] )

    for a,b in edges:
      nodes_to_added_edges.add(a)
      nodes_to_added_edges.add(b)

    ### decrease edges that would be eliminated for node a by the
    ### number of edges that have actually been added this
    ### iteration (note that it is faster to do this in a batch
    ### manner because each decrease_key operation is log(n))
    for a in nodes_to_added_edges:
      x = nodes_to_heap_references[a]
      #print('\tchanging',a,':',x.ref.key,'+',nodes_to_added_edges[a],'- 1')
      ### store the old number of neighbors
      old_num_neighbors = x.ref.key
      ### remove the node by decreasing its key lower than any other
      x.delete()
      nodes_to_heap_references[a] = bh.insert(old_num_neighbors + nodes_to_added_edges[a] - 1,a)
  
  local.add_edges_from(edges)
  added_edges.extend(edges)
  local.remove_node(n)

  #to_min = [ (len(local[n]), n) for n in local ]
  #print(str(to_min))
  #if len(to_min) > 0:
  #  print('min',min(to_min))

  result = g.copy()
  return result

### does minimum fill-in
def connected_to_triangulated(g):
  added_edges = []
  local = g.copy()
  while len(local) > 0:
    ### count the edges that would be added by each node
    ### elimination; eliminate the node that will result in the
    ### fewest
    number_of_added_edges_for_nodes = [ (number_of_added_elimination_edges(local,n), n) for n in local ]
    number_of_edges, n = min(number_of_added_edges_for_nodes)
    #print('removing',n,'with',number_of_edges,'edges')

    edges = elimination_edges(local, n)
    local.add_edges_from(edges)
    added_edges.extend(edges)
    local.remove_node(n)
  result = g.copy()
  result.add_edges_from(added_edges)
  return result

def to_clique_graph(g):
  cliques = [ c for c in nx.find_cliques(g) ]
  clique_graph = nx.make_max_clique_graph(g)
  ### networkx.clique_graph does not tell which nodes it contains
  
  frozenset_clique_graph = nx.Graph()
  for a in clique_graph:
    a_clique = frozenset(cliques[a-1])
    frozenset_clique_graph.add_node(a_clique)
    for b in clique_graph[a]:
      b_clique = frozenset(cliques[b-1])
      ### weight it with the negative cardinality of the
      ### intersection so that you can use the built in
      ### minimum_spanning_tree function to get the maximum
      ### spanning tree
      overlap_size = len(a_clique.intersection(b_clique))
      frozenset_clique_graph.add_edge( a_clique, b_clique, weight = -overlap_size)

  return frozenset_clique_graph

def moralized_to_junction_tree(moral_graph):
  junction_tree = nx.Graph()
  #print('\t\ttriangulating...')
  triangulated = to_triangulated(moral_graph)
  #print)'\t\tbuilding clique graph...')
  clique_graph = to_clique_graph(triangulated)
  #print('\t\tfinding max spanning tree...')
  junction_tree = nx.minimum_spanning_tree(clique_graph)
  return junction_tree

def verify_junction_forest(junction_forest):
  for jt in nx.connected_component_subgraphs(junction_forest):
      verify_junction_tree(jt)

def verify_junction_tree(junction_tree):
  ### uses my own method for verification

  ### Cut each edge L--R. The tree is only a valid junction tree if the
  ### set of nodes in the left tree does not contain the nodes removed
  ### by moving from R to L (and vice-versa).

  ### Intuitive proof: Each running intersection of any element must
  ### have a boundary (or is ubiquitous throughout the tree). Past the
  ### boundary, the element cannot be featured.
  local_jt = junction_tree.copy()
  for sL, sR in junction_tree.edges():
    local_jt.remove_edge(sL,sR)
    subtrees = nx.connected_component_subgraphs(local_jt)

    ### if this is a proper tree, then it should have only two
    ### connected subgraphs
    if len(subtrees) != 2:
      raise Exception('Error: verify junction tree was not called using a connected tree')
    
    subtreeL = subtrees[0]
    subtreeR = subtrees[1]
    if sR in subtreeL:
      subtreeL, subtreeR = subtreeR, subtreeL

    newR = sR.difference(sL)
    newL = sL.difference(sR)
    if len(newR.intersection(subtreeL)) > 0 or len(newL.intersection(subtreeR)) > 0:
      print('verify_junction_tree: cutting edge',sL,'--',sR,'results in overlapping edges')
      return
    local_jt.add_edge(sL,sR)

def graph_union(g1, g2):
  if g1 == None:
    return g2
  if g2 == None:
    return g1
  for n in g2:
    g1.add_node(n, **g2.node[n])
    for m in g2[n]:
      g1.add_edge(n,m,**g2[n][m])
  return g1

def all_graph_unions(graph_lst):
  ### do not set as Graph or DiGraph so that it works with any
  ### homogeneous list
  result = None
  for g in graph_lst:
    result = graph_union(result, g)
  return result

def treewidth(junction_tree):
  return max([len(fs) for fs in junction_tree])

def type_subgraph(g, t):
  type_nodes = [ n for n in g if n['type'] == t ]
  return g.subgraph(type_nodes)

def log_connected_naive_complexity(dg):
  return len(type_subgraph(dg, 'protein'))

def log_connected_naive_junction_tree_complexity(junction_tree):
  var_sets = [fs for fs in junction_tree]
  res = set()
  for vs in var_sets:
    res.update(vs)
  return len(res)

def log_junction_tree_complexity(junction_tree):
  return log_sum([ len(fs) for fs in junction_tree ])




# coding: utf-8

import logging

logger = logging.getLogger('root')

class ConfigFileError(Exception):
  def __init__(self,*args,**kwargs):
    logger.error(str(*args))
    super().__init__(*args,**kwargs)

class FilteringError(Exception):
  def __init__(self,*args,**kwargs):
    logger.error(str(*args))
    super().__init__(*args,**kwargs)

class STANError(Exception):
  def __init__(self,*args,**kwargs):
    logger.error(str(*args))
    super().__init__(*args,**kwargs)


# coding: utf-8

import argparse
import logging
import numpy as np
import os
import pandas as pd
import pkg_resources
import sys
from yaml import load as yaml_load, dump as yaml_dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


from dart_id.exceptions import ConfigFileError
from dart_id.version import __version__
from shutil import copyfile

logger = logging.getLogger('root')

# set functions taken from https://www.saltycrane.com/blog/2008/01/how-to-find-intersection-and-union-of/

#def unique(a):
#    """ return the list with duplicate elements removed """
#    return list(set(a))

def intersect(a, b):
    """ return the intersection of two lists """
    return list(set(a) & set(b))

def union(a, b):
    """ return the union of two lists """
    return list(set(a) | set(b))

def init_logger(verbose, log_file_path, log_to_file=True):
  # set up logger
  for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler) 
   
  logFormatter = logging.Formatter('%(asctime)s [%(threadName)-5.5s] [%(levelname)-8.8s]  %(message)s')
  logger = logging.getLogger('root')

  if   verbose == 3: logger.setLevel(logging.DEBUG)
  elif verbose == 2: logger.setLevel(logging.INFO)
  elif verbose == 1: logger.setLevel(logging.WARNING)
  elif verbose == 0: logger.setLevel(logging.ERROR)
  else: logger.setLevel(logging.WARNING) # default

  if log_to_file:
    fileHandler = logging.FileHandler(log_file_path, mode='w')
    fileHandler.setFormatter(logFormatter)
    logger.addHandler(fileHandler)

  consoleHandler = logging.StreamHandler()
  consoleHandler.setFormatter(logFormatter)
  logger.addHandler(consoleHandler)

  logger.info(' '.join(sys.argv[0:]))

  return logger

def create_fig_folder(output_path, fig_folder):
  figures_path = os.path.join(output_path, fig_folder)
  if not os.path.exists(figures_path):
    logger.info('Path for figures folder {} does not exist. Creating...'.format(figures_path))
    os.makedirs(figures_path)
  return figures_path

def load_params_from_file(params_folder):
  # first expand user or any vars
  params_folder = os.path.expanduser(params_folder)
  params_folder = os.path.expandvars(params_folder)

  # load parameters if they are specified in the command line
  params = {}
  logger.info('Using provided alignment parameters. Loading params from {}...'.format(params_folder))
  param_files = ['exp_params.txt', 'pair_params.txt', 'peptide_params.txt']
  for pf in param_files:
    pfp = os.path.join(params_folder, pf)
    if os.path.exists(pfp):
      try:
        params[pf.split('_')[0]] = pd.read_csv(pfp, sep='\t')
      except:
        logger.error('Error loading param file')
    else:
      raise ConfigFileError('Params file {} does not exist'.format(pfp))

    logger.info('Loaded \"{}\" params file.'.format(pf.split('_')[0]))

  return params

def add_global_args(parser, add_config_file=True):
  parser.add_argument('-i', '--input', type=argparse.FileType('r'), nargs='+', default=None, help='Input file(s) from search engine output (e.g., MaxQuant evidence.txt). Not required if input files are specified in the config file')
  #parser.add_argument('-o', '--output', help='Path to converted file. Default: prints to stdout')
  parser.add_argument('-o', '--output', type=str, default=None, help='Path to output folder')
  parser.add_argument('-v', '--verbose', type=int, default=1, help='Level of logging. 0 = ERROR, 1 = WARNING (default), 2 = INFO, 3 = DEBUG')
  parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__), help='Display the program\'s version')

  if add_config_file:
    parser.add_argument('-c', '--config-file', required=True, type=argparse.FileType('r', encoding='UTF-8'), help='Path to config file (required). See example/config_example.yaml')

def read_default_config_file():
  # load input file types
  default_config = pkg_resources.resource_stream('dart_id', '/'.join(('config', 'defaults.yaml')))
  default_config = yaml_load(default_config, Loader=Loader)
  return default_config

def read_config_file(args, create_output_folder=True):
  # load defaults
  config = read_default_config_file()

  # override defaults with user config file
  with open(args.config_file.name, 'r') as f:
    config.update(yaml_load(f, Loader=Loader))
    
  # override config file's input, output, and verbose options
  # if they were specified on the command-line
  if args.input is not None:
    if config['input'] is not None:
      logger.warning('Input files specified in both the config file and the command line. Using command-line input files instead.')
    config['input'] = [f.name for f in args.input]

  if args.output is not None:
    if 'output' in config and config['output'] is not None:
      logger.warning('Output folder specified in both the config file and the command line. Using command-line output folder instead.')
    config['output'] = args.output

  if args.verbose:
    if 'verbose' in config:
      logger.info('Overwriting verbosity level in configuration file with the one provided on the command-line.')
      config['verbose'] = args.verbose

  # make sure that we have inputs and outputs before continuing
  if config['input'] is None:
    raise ConfigFileError('No input files specified, in either the config file or the command line. Please provide input files.')

  if 'output' not in config or config['output'] is None:
    raise ConfigFileError('No output folder specified, in either the config file or the command line. Please provide output folder.')

  ### --------------------
  ### VALIDATE CONFIG FILE
  ### --------------------

  ### VALIDATE FILTERS
  filters = config['filters']
  
  # each filter has a specified required column from the dataframe
  # make sure these columns exist before proceeding
  for i, f in enumerate(filters):
    # make sure filter has required field of 'name'
    if 'name' not in f or f['name'] is None or len(f['name']) == 0:
      raise ConfigFileError('Filter {} does not have the required field name or it is empty. Please provide a name (type) for this filter. Available filters are listed in the example config file, under example/config.yaml'.format(str(f)))

  # expand user or any vars
  config['output'] = os.path.expanduser(config['output'])
  config['output'] = os.path.expandvars(config['output'])

  # create output folder
  if not os.path.exists(config['output']) and create_output_folder:
    logger.info('Output folder does not yet exist. Creating...')
    os.makedirs(config['output'])

  # copy config file to output folder
  if create_output_folder:
    logger.info('Copying config file to output folder')
    copyfile(args.config_file.name, os.path.join(config['output'], os.path.basename(args.config_file.name)))

  return config

# TODO: do this dumb input check every time the config file is loaded
def check_config_inputs(config):
  pass

# pulled from: https://stackoverflow.com/a/29677616
def weighted_quantile(values, quantiles, sample_weight=None, values_sorted=False, old_style=False):
  """ Very close to numpy.percentile, but supports weights.
  NOTE: quantiles should be in [0, 1]!
  :param values: numpy.array with data
  :param quantiles: array-like with many quantiles needed
  :param sample_weight: array-like of the same length as `array`
  :param values_sorted: bool, if True, then will avoid sorting of initial array
  :param old_style: if True, will correct output to be consistent with numpy.percentile.
  :return: numpy.array with computed quantiles.
  """
  
  values = np.array(values)
  quantiles = np.array(quantiles)

  if sample_weight is None:
    sample_weight = np.ones(len(values))
  sample_weight = np.array(sample_weight)
  assert np.all(quantiles >= 0) and np.all(quantiles <= 1), 'quantiles should be in [0, 1]'

  if not values_sorted:
    sorter = np.argsort(values)
    values = values[sorter]
    sample_weight = sample_weight[sorter]

  weighted_quantiles = np.cumsum(sample_weight) - 0.5 * sample_weight
  if old_style:
    # To be convenient with np.percentile
    weighted_quantiles -= weighted_quantiles[0]
    weighted_quantiles /= weighted_quantiles[-1]
  else:
    weighted_quantiles /= np.sum(sample_weight)
  return np.interp(quantiles, weighted_quantiles, values)

def weighted_median(values, weights):
  order = np.argsort(values)
  values = values[order]
  weights = weights[order]

  #weighted_quantiles = (np.cumsum(weights) - (0.5 * weights)) / np.sum(weights)
  
  return np.interp([0.5], 
    (np.cumsum(weights) - (0.5 * weights)) / np.sum(weights),
    values)
    
def convert_numpy_scalar(x):
  if type(x) == float or type(x) == int:
    return x
  else: return x.tolist()

